/* PM 2017
 * Radu Ioana-Roxana
 * 
 * Cod sursa pentru proiect: telecomanda care transmite comenzi prin bluetooth unei masinute.
 *
 */

#define F_CPU 16000000

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdlib.h>
#include <math.h>

#include "usart.h"
#include "lsm6.h"

#define MAX 16000

#define ON 1
#define OFF 0

#define DEFAULT 'N'
#define MOVE_FRONT 'F'
#define MOVE_BACK 'B'
#define MOVE_RIGHT 'R'
#define MOVE_LEFT 'L'
#define BUZZER_START 'Z'
#define BUZZER_STOP 'Y'
#define LIGHTS_ON 'U'
#define LIGHTS_OFF 'V'

/* prag pentru inclinatia accelerometrului */
int treshold = 3000;

/* valori standard pentru accelerometru */
int xstd = -210;
int ystd = 500;

/* starea butonului*/
uint8_t button_pressed = 0;
/* starea farurilor*/
uint8_t lights_state = OFF;
int k = 0;

/* Functie ce initializeaza ledurile*/
void init_leds()
{	
	/* setare PA0 ca iesire */
	DDRA |= (1 << PA0);
	
	/* setare PA1 ca iesire */
	DDRA |= (1 << PA1);
	
	/* setare PA2 ca iesire */
	DDRA |= (1 << PA2);
	
	/* setare PA3 ca iesire */
	DDRA |= (1 << PA3);
}

/* Functie ce initializeaza butonul */
void init_button()
{
	/* Setez pinul PA4 ca intrare. */
	DDRA &= ~(1 << PA4);

	/* Activez rezistenta de pull-up pentru pinul PA4 */
	PORTA |= (1 << PA4);
	
	/* intrerupere activata pentru PCINT4 (PA4) */
	PCICR  = (1 << PCIE0);
    PCMSK0 = (1 << PCINT4);
}

/* Functie de initializare a timerului 1 pt a genera intrerupere la 1 secunda
 * Am folosit calculatorul din laborator
 */
void init_timer1()
{
    
    /* activeaza intreruperea */
    TIMSK1 = (1 << OCIE1A);
	
	OCR1A = 62500;
    
    /*mod ctc standard */
    TCCR1B = (1 << CS12);
}

/* Aprinde ledul care indica directia fata */
void turnOn_front()
{
	/* turn on PA1 */
	PORTA &= ~(1 << PA1);
	PORTA |= (1 << PA1);
}

/* Aprinde ledul care indica directia spate */
void turnOn_back()
{
	/* turn on PA2 */
	PORTA &= ~(1 << PA2);
	PORTA |= (1 << PA2);
}

/* Aprinde ledul care indica viraj stanga */
void turnOn_left()
{
	/* turn on PA0 */
	PORTA &= ~(1 << PA0);
	PORTA |= (1 << PA0);
}

/* Aprinde ledul care indica viraj dreapta */
void turnOn_right()
{
	/* turn on PA3 */
	PORTA &= ~(1 << PA3);
	PORTA |= (1 << PA3);
}

/* Stinge toate ledurile */
void turnOff_all()
{
	PORTA &= ~(1 << PA0);
	PORTA &= ~(1 << PA1);
	PORTA &= ~(1 << PA2);
	PORTA &= ~(1 << PA3);
}

/* Obtine o valoare pe 2 octeti din 2 octeti: unul de high si unul de low */
short make_value(uint8_t low, uint8_t high)
{
	short val = low;
	val |= (high << 8);
	
	return val;
}

/* Intreruperea de schimbare a starii butonului */
ISR (PCINT0_vect)
{
	if ((PINA & (1 << PA4)) == 0) {
		if (!button_pressed)
			button_pressed = 1;
	}
	else {
		if(button_pressed)
			button_pressed = 0;
	}
	
	_delay_ms(10);
}

/* Intreruperea de comparatie cu OCR1A pentru timerul 1 */
ISR(TIMER1_COMPA_vect)
{
    k = 0;
}

/* Obtinem viteza in functie de inclinatia accelerometrului
 * Conform protocolului 
 * 			Pentru valori pana in prag viteza va fi 0 iar valoarea dorita va fi 255.
 * 			Pentru inclinatie maxima viteza va fi maxima iar valoarea dorita este 0.
 * Am folosit o functie de gradul 1 f(x) = ax + b, care pentru prag va fi >= 255 si pentru MAX va fi 0.
 * Deoarece la trimiterea prin usart caracterele cu cod ascii <= 10 se comportau "ciudat" (unele erau ignorate)
 * am ales ca pentru valori obtinute mai mici ca 10 acestea sa fie scalate la 10.
 */
uint8_t scale_speed(short inclination)
{
	float speed;
	
	if (abs(inclination) < treshold)
		return 255;
	
	if ( abs(inclination) >= MAX)
		return 10;
		
	speed = -0.02 * abs(inclination) + 320;
	
	if ( floor(speed) >= 255)
		return 255;
	
	if( floor(speed) <= 1)
		return 1;
	
	uint8_t ret_speed = (uint8_t)speed;
	
	if (ret_speed <= 10) {
		return 10;
	}
	
	return ret_speed;
}

void run()
{
	uint8_t accX_low, accX_high, accY_low, accY_high, accZ_low, accZ_high;
	uint8_t gyrX_low, gyrX_high, gyrY_low, gyrY_high, gyrZ_low, gyrZ_high;

	short accX = treshold - 1, accY, accZ, gyrX, gyrY, gyrZ;
	int front, back, right, left;
	int i;
	uint8_t speed;
	
	while(1) {
	
		/* contorizam cate rezultate de miscare fata/spata stanga/dreapta am primit */
		front = 0;
		back = 0;
		left = 0;
		right = 0;
		
		int ok1 = 0, ok2 = 0;
		
		/* Citim acceleromtrul de 2000 de ori la interval de 5 microsecunde pentru a asigura 
		 * consistenta datelor si a incerca sa eliminam spike-urile/erorile */
		for (i = 0; i < 2000; i++) {
		
			_delay_us(5);
			read_acc(&accX_low, &accX_high, &accY_low, &accY_high, &accZ_low, &accZ_high);
			
			accX = make_value(accX_low, accX_high);
			accY = make_value(accY_low, accY_high);
			accZ = make_value(accZ_low, accZ_high);
			
			if ( abs(xstd - accX) > treshold) {
				if ( xstd - accX > 0)
					front += 1; /* am obtinut miscare in fata*/
				else
					back += 1; /* am obtinut miscare in spate */
			}
			
			if ( abs(ystd - accY) > treshold) {
				if( (ystd - accY) > 0)
					right += 1; /* am obtinut viraj stanga*/
				else
					left += 1; /* am obtinut viraj dreapta */
			}		
		}
		
		/* citim giroscopul */
		read_gyro(&gyrX_low, &gyrX_high, &gyrY_low, &gyrY_high, &gyrZ_low, &gyrZ_high);
		gyrX = make_value(gyrX_low, gyrX_high);
		gyrY = make_value(gyrY_low, gyrY_high);
		gyrZ = make_value(gyrZ_low, gyrZ_high);
		
		/* daca avem micsare pe giroscop, ignoram accelerometrul */
		if ( abs(gyrZ) > 22000 && k == 0) {
			if (lights_state == ON)
				lights_state = OFF;
			else
				lights_state = ON;

			/* resetare timer */
			TCNT1H = 0;
			TCNT1L = 0;
			
			/* setam k 1 deoarece ne dorim ca giroscopul sa fie citit la fiecare 1 secunda */
			k = 1; /* k va deveni 0 dupa o secunda, in ISR-ul corespunzator */

			/* ignoram accelerometrul prin setarea statisticilor precedente ca fiind inconsistente */
			/* deoarece la rotatia pe ox se poate declansa accelerometrul cu toate ca intentia nu era aceasta */
			back = front = left = right = 1;
		}
		
		/* stingem toate ledurile pentru a le reaprinde doar pe cele corespunzatoare la acest moment de timp */
		turnOff_all();
		
		/* verificare consistenta a rezultatelor */
		if ( (back == 2000 && front == 0) || (back == 0 && front == 2000 )|| (front == 0 && back == 0))
			ok1 = 1;
		
		if ( (left == 2000 && right == 0) || (left == 0 && right == 2000) || (left == 0 && right == 0))
			ok2 = 1;
		
		/* Bufferul de trimitere [Fata/spate][stanga/dreapta][viteza][stare_claxon][stare_faruri] */
		char buf_to_send[6];
		
		for (i = 0; i < 5; i++)
			buf_to_send[i] = DEFAULT;
		buf_to_send[5] = '\0';
		
		/* obtinem viteza in functie de inclinatie */
		speed = scale_speed(accX);

		/* daca avem date consistente completam bufferul de trimitere corespunzator */
		if(ok1 && ok2) {
			if (front) {
				turnOn_front();
				buf_to_send[0] = MOVE_FRONT;
			}
			else
				if(back) {
					turnOn_back();
					buf_to_send[0] = MOVE_BACK;
				}
					
			if(left) {
				turnOn_left();
				buf_to_send[1] = MOVE_LEFT;
			}
			else
				if(right) {
					turnOn_right();
					buf_to_send[1] = MOVE_RIGHT;
				}
		}
		
		buf_to_send[2] = speed;
		buf_to_send[3] = button_pressed == ON ? BUZZER_START : BUZZER_STOP;
		buf_to_send[4] = lights_state == ON ? LIGHTS_ON : LIGHTS_OFF;
		buf_to_send[5] = '\0';
		USART0_send_string(buf_to_send);
	}
}

/* Confiruatiile inainte de pornire */
void configure(void)
{
	/* configuratii initiale */	
	sei();
	USART0_init();
	init_leds();
	init_lsm6();
	init_button();
	init_timer1();
}


int main(void) {

	configure();
	run();
	return 0;
}
